from . import godot_script_1
from . import godot_script_2
from . import csharp
from . import resource
from . import json

registers = {}

registers['GDS1.0'] = godot_script_1
registers['GDS2.0'] = godot_script_2
registers['C#'] = csharp
registers['RESOURCE'] = resource
registers['JSON1.0'] = json
registers['JSON2.0'] = json



import glob
import os
import pprint
import logging
import textwrap
from excelexporter.config import Configuration
from excelexporter.sheetdata import SheetData
from excelexporter.generator import Converter, Type

# 导出格式
extension = "gd"

logger = logging.getLogger(__name__)

func_names = []
func_codes = []

cvt = Converter()

cvt.register(Type.STRING, lambda v, n, id, p: str(v) if v else "")
cvt.register(
    Type.INT, lambda v, n, id, p: int(str(v or 0).split(".")[0])
)
cvt.register(Type.FLOAT, lambda v, n, id, p: float(str(v or 0)))
cvt.register(Type.BOOL, lambda v, n, id, p: v != "FALSE")
cvt.register(
    Type.ARRAY,
    lambda v, n, id, p: eval(f'[{v.replace("|",",")}]') if v else []
)
cvt.register(
    Type.ARRAY_STR,
    lambda v, n, id, p: ["%s" %
                         e for e in v.split("|")]if v else []
)
cvt.register(
    Type.ARRAY_BOOL,
    lambda v, n, id, p: [e != "FALSE" for e in v.split("|")] if v else []
)
cvt.register(
    Type.DICT,
    lambda v, n, id, p: eval(f'{{{v.replace("|",",")}}}')
    if v else {}
)
cvt.register(
    Type.FUNCTION,
    lambda v, n, id, p: make_func(v or "pass", n, id, p)
)


def make_func(v, fn, id, p):
    func_name = f"{fn}_{id}"

    params = "(args=[])" if not p else p

    func_code = textwrap.dedent(
        f"""
static func {func_name}{params}:
{textwrap.indent(v, "    ")}
"""
    )
    func_names.append(func_name)
    func_codes.append(func_code)
    return func_name


def generator(sheetdata: SheetData, config: Configuration):
    # 表格数据脚本模板
    template = """
    extends Reference
    var None = null
    var False = false
    var True = true

    var data = \\
    {data}

    {funcs}
    """
    template = textwrap.dedent(template)
    field_names = sheetdata.define.name
    field_types = sheetdata.define.type
    table = {}

    for row in sheetdata.table:
        id_type = field_types[0]
        id = cvt(id_type, row[0], None, None)

        row_data = {}

        for index, value in enumerate(row):
            field_name: str = field_names[index]
            field_type = field_types[index]
            row_data[field_name] = cvt(field_type, value, field_name, id)

        table[row_data["id"]] = row_data

    funcs = textwrap.dedent("\n".join(func_codes))

    code = template.format(
        data=pprint.pformat(table, indent=2, width=1000000000,
                            compact=True, sort_dicts=True),
        funcs=funcs
    )

    code = textwrap.dedent(code)

    for func_name in func_names:
        code = code.replace(
            f"'{func_name}'",
            f"funcref(self,'{func_name}')"
        )

    return code


def completed_hook(config: Configuration):

    output = config.output
    settings_file_path = os.path.join(output, "settings.gd")
    project_root = config.project_root

    lines = []

    for path in glob.glob(f"{output}/**/*.*", recursive=True):
        if path == settings_file_path:
            continue  # 跳过 settings.gd
        basename = os.path.basename(path)
        setting_name = os.path.splitext(basename)[0]
        relpath = os.path.relpath(path, project_root).replace("\\", "/")
        lines.append(f"var {setting_name} = load('res://{relpath}').new()")

        # 去掉缩进
    code = textwrap.dedent(
        """
    extends Node
    # 这个脚本你需要挂到游戏的Autoload才能全局读表

    {refs_code}
    """
    )
    refs_code = "\n".join(lines)

    code = code.format(refs_code=refs_code)

    with open(settings_file_path, "w", encoding="utf-8", newline="\n") as f:
        f.write(code)
        logger.info("创建setting.gd")



import logging

from .sheetdata import SheetData
from .config import Configuration
from typing import Any, Callable


Generator = Callable[[SheetData, Configuration], str]
CompletedHook = Callable[[Configuration], None]

ConvertFunc = Callable[[Any, str, int, dict], str]

logger = logging.getLogger()


class Type:
    STRING = "string"
    INT = "int"
    FLOAT = "float"
    BOOL = "bool"
    ARRAY = "array"
    ARRAY_STR = "array_str"
    ARRAY_BOOL = "array_bool"
    DICT = "dict"
    FUNCTION = "function"


class Converter:

    def __init__(self) -> None:
        self._map = {}

    def default(self, v, fn, id, params): return v or 0

    def register(self, type: str, cvt_method: ConvertFunc):
        self._map[type] = cvt_method

    def __call__(
            self,
            type: str,
            value: Any,
            field_name: str,
            id: int,
            *args: Any,
            **kwds: Any
    ) -> Any:
        type = type.strip()
        type_name, params = type.split("#") if "#" in type else (type, None)
        cvt = self._map.get(type_name, self.default)
        result = cvt(value, field_name, id, params)
        return result



from dataclasses import dataclass, field


@dataclass
class Define:
    type: list[str] = field(default_factory=list)
    desc: list[str] = field(default_factory=list)
    name: list[str] = field(default_factory=list)


@dataclass
class SheetData:
    define: Define = field(default_factory=Define)
    table: list = field(default=list)



import glob
import os
from typing import Optional
import xlwings as xw
import logging

from excelexporter.sheetdata import SheetData
from .generator import Generator, CompletedHook
from .config import Configuration
from .generators import registers

# 导表工具引擎

logger = logging.getLogger(__name__)


class IllegalFile(Exception):

    def __init__(self, filename: str, *args: object) -> None:
        super().__init__(f"{filename} 不是配置表目录下的配置", *args)


class IllegalGenerator(Exception):
    def __init__(self, name: str, *args: object) -> None:
        super().__init__(
            name,
            *args
        )


class Engine(xw.App):

    def __init__(self, config: Configuration) -> None:
        super().__init__(visible=False)
        self.config = config
        self.generator: Optional[Generator] = None
        self.completed_hook: Optional[CompletedHook] = None
        self.extension: str = ""

        self.init_generator()

    def init_generator(self):

        generator = None
        completed_hook = None
        extension = None

        # 如果有自定义导出器就优先用自定义导出器
        if self.config.custom_generator.endswith(".py"):
            with open(self.config.custom_generator) as f:
                code = f.read()
                exec(code)
                self.generator = generator
                self.completed_hook = completed_hook
                self.extension = extension

                if (generator and completed_hook and extension) is False:
                    raise IllegalGenerator(
                        self.config.custom_generator,
                        "自定义导出器不完整，generator、completed_hook、extension存在没定义。")

                logger.info(f"使用 {self.config.custom_generator} 自定义导出器")
        else:
            # 没有才用内置的
            module = registers.get(
                self.config.custom_generator.upper()
            )
            if module is None:
                raise IllegalGenerator(
                    self.config.custom_generator, "不是内置导出器，请检查配置。")
            logger.info(
                f"使用内置导出器 {self.config.custom_generator} :{module.__name__}")
            self.generator = getattr(module, "generator")
            self.completed_hook = getattr(module, "completed_hook")
            self.extension = getattr(module, "extension")

    def _gen(self, excel_file: str):
        wb_abs_path: str = os.path.abspath(excel_file)
        abs_input_path: str = os.path.abspath(self.config.input)
        abs_output_path: str = os.path.abspath(self.config.output)
        wb_abs_path_without_ext: str = os.path.splitext(wb_abs_path)[0]

        if not wb_abs_path.startswith(abs_input_path):
            raise IllegalFile(wb_abs_path, abs_input_path)

        sheet_datas = self._excel2dict(wb_abs_path)

        for name, sheetdata in sheet_datas.items():
            try:
                if "-" in name:
                    org_name, rename = name.split("-")
                else:
                    org_name, rename = name, None

                relative_path = os.path.join(
                    wb_abs_path_without_ext.replace(abs_input_path, ""),
                    rename or org_name
                )
                output = abs_output_path + relative_path

                output_dirname = os.path.dirname(output)
                # 保持输出的文件目录层级结构与输入一致
                if not os.path.exists(output_dirname):
                    os.makedirs(output_dirname)

                code = self.generator(sheetdata, self.config)

                # code = "# 本文件由代码生成，不要手动修改\n"+code
                output = f"{output}.{self.extension}"
                with open(output, "w", encoding="utf-8", newline="\n") as f:
                    f.write(code)
                    logger.info(f"导出：{wb_abs_path}:{name} => {output}")
            except Exception:
                logger.error(f"{name} 导出失败", exc_info=True)

    def _excel2dict(self, wb_file: str) -> dict[str, SheetData]:
        """
        workbook解析加工成字典
        """
        book = self.books.open(wb_file)

        ignore_sheet_mark = self.config.ignore_sheet_mark
        # 过滤掉打了忽略标志的sheet
        sheets = filter(
            lambda sheet: not sheet.name.startswith(ignore_sheet_mark),
            book.sheets
        )

        sheet_data = {}

        for sheet in sheets:

            data = SheetData()

            row_values = sheet.range("A1").expand().raw_value

            data.define.type = list(row_values[0])
            data.define.desc = list(row_values[1])
            data.define.name = list(row_values[2])

            data.table = list([list(row) for row in row_values[3:]])

            # 找出所有被打了忽略标记的字段
            for col, field in enumerate(data.define.name):
                if field.startswith(self.config.ignore_field_mark):
                    del data.define.type[col]
                    del data.define.desc[col]
                    del data.define.name[col]
                    for row in data.table:
                        del row[col]

            sheet_data[sheet.name] = data

        return sheet_data

    def gen_one(self, filename: str):
        self._gen(filename)
        if self.completed_hook:
            self.completed_hook(self.config)

    def gen_all(self):
        abs_input = os.path.abspath(self.config.input)
        exts = [".xlsx", ".xls"]
        for ext in exts:
            full_paths = glob.glob(f"{abs_input}/**/*{ext}", recursive=True)
            for full_path in full_paths:
                filename = os.path.basename(full_path)
                if filename.startswith("~$"):
                    logger.warning(f"{filename} 不是配置表，跳过！")
                    continue
                self._gen(full_path)

        if self.completed_hook:
            self.completed_hook(self.config)
